#ifndef ALIGNER_HH
#define ALIGNER_HH

#include "base.hh"
#include "sequence.hh"
#include "basepairs.hh"
#include "alignment.hh"

#include "params.hh"

#include "matrices.hh"

/*

TODO: score matrices should be as small as possible
      and have offsets.
      The M-Matrix can be smaller if the maximal arc len is
      limited. The D-matrix may be smaller if the number
      of simultaneously needed arc-pairs is limited, due
      to limit on the local sub-sequence lengths.

      A special cyclically rotatable matrix is needed for
      alignment on the top-level

*/


/**
   Contains data for restricting Aligner to sub-sequences

   take care when using aligner restrictions for multiple
   Alignments with the same aligner object.
   The D-matrix is only computed once, so this works
   as long as the first Aligner::align() is called with
   the most general restriction (e.g. no restriction at all)!
*/

class AlignerRestriction {
private:
    int startA;
    int startB;
    int endA;
    int endB;
public:
    AlignerRestriction(int startA_, int startB_, int endA_, int endB_)
	: startA(startA_), startB(startB_), endA(endA_), endB(endB_)
    {}

    uint get_startA() const {return startA;}
    uint get_endA() const {return endA;}
    uint get_startB() const {return startB;}
    uint get_endB() const {return endB;}

    void set_startA(uint p) {startA=p;}
    void set_endA(uint p)  {endA=p;}
    void set_startB(uint p)  {startB=p;}
    void set_endB(uint p)  {endB=p;}
};

inline
std::ostream & operator<<(std::ostream &out,AlignerRestriction r) {
    return 
	out << r.get_startA() << " "
	    << r.get_startB() << " "
	    << r.get_endA() << " "
	    << r.get_endB();
}



typedef Matrix<score_t> ScoreMatrix;


/* ============================================================ 
   class Aligner

   Compute alignment similarity score and optimal alignment
   
   Methods:
   - Aligner
   initialize with two sequences and their basepairs (including scores)
   - sigma
   similartiy for bases
   - tau 
   sequence contribution for arc-matches
   - ...
   ============================================================ */

#define FOR_IJ_IN_ARCMATCH \
  for (int i=al+1; i<ar; i++) \
    for (int j=bl+1; j<br; j++)

//! class for doing the alignment of two sequences and their
//! associated sets of weighted basepairs
//!
//! an object always knows about the two sequences and
//! the two weighted base pair sets
//!
//! common usage: construct, align, trace, get_alignment
class Aligner {
protected:
    AlignerParams *params; //!< the parameter for the alignment
	
    // The following is added for the idea of 
    // getting the k-best local alignments
    //
    // Aligner always works on sub-sequences/structures of
    // seqA and seqB. These are specified by
    AlignerRestriction r;
    // in the standard case, i.e. align the whole sequences,
    // the values are set to startA=1 and endA=seqA.length()
    // and analogously for B
	
    // the constructor initializes to the above standard values
    // methods provide possibility to restriction

protected:
    Sequence &seqA;
    BasePairs &bpsA;
    Sequence &seqB;
    BasePairs &bpsB;
  
    ScoreMatrix D;
    std::vector<ScoreMatrix> Ms;
  
    Alignment alignment;
  
    Matrix<int> sigma_tab;
  
    int max_i,max_j; // computed by align_top_level

    bool D_created; // flag, is D already created?

    enum {E_NO_NO, E_X_NO, E_NO_X, E_X_X,
	  E_OP_NO, E_NO_OP, E_OP_X, E_X_OP};
  
    //! compute base similarity
    int sigma_(int i, int j);

    //! precompute all base similarities
    //! and put them in the sigma table
    void precompute_sigma();  

    //! return the (precomputed) base similarity
    int sigma(int i, int j) {return sigma_tab(i,j);}

    //! contribution of sequence match/mismatch to the arc match score
    int tau(int i, int j, int k, int l) {
	return (params->tau_factor)?( ((sigma(i,j)+sigma(k,l)) * params->tau_factor) / 100 ):0;
    }

    //! initialize first column and row
    void init_state(ScoreMatrix &M, int al, int ar, int bl, int br, 
		    score_t indelA, score_t exclusion_scoreA,
		    score_t indelB, score_t exclusion_scoreB);
    
    // recursion cases that handle everything but exclusions
    // (in the LSSA-paper this function was called NoEx
    
    //! standard cases for alignment (without exlusion handling)
    score_t align_noex(ScoreMatrix &M,int al,int bl,int i, int j);
    
    //! arc match handling in noLP-mode
    score_t align_noex_noLP(score_t max_score,ScoreMatrix &M,int al,int bl,int i, int j);
 
    //! align, where al,ar,bl,br is an arc-match
    //! in structure local alignment, this allows to introduce exclusions
    void align_in_arcmatch(int al,int ar,int bl,int br,
			   bool allow_exclusion);
  
    //! align the top-level in a sequence local alignment
    //! and return the maximal score
    int align_top_level_locally();

    //! align top level in the scanning version
    int align_top_level_localB();
  
    //! trace back within an match of arcs
    void trace_in_arcmatch(int state,int al,int i,int bl,int j,bool top_level);
  
    //! standard cases in trace back (without handling of exclusions)
    void trace_noex(int state,int al,int i,int bl,int j,bool top_level);

    /**
       create the entries in the D matrix
       This function is called by align() (unless D_created)
    */
    void align_D();

    /**
       fill in D the entries with left ends al,bl, 
       where adjlA, adjlB are adjanceny lists of the arcs
    */
    void 
    fill_D_entries(int al, int bl,
		   const BasePairs::LeftAdjList &adjlA,
		   const BasePairs::LeftAdjList &adjlB,
		   int max_ar, int max_br
		   );

public:  
    //! copy constructor
    Aligner(const Aligner &a);

    //! construct with sequences and corresponding sets of basepairs
    Aligner(Sequence &_seqA, BasePairs &_bpsA,
	    Sequence &_seqB, BasePairs &_bpsB,
	    AlignerParams *_ap
	    );
    
    //! return the alignment that was computed by trace()
    Alignment const & get_alignment() const {return alignment;} 
    
    //! assignment operator
    const Aligner & operator = (const Aligner &a);
    
    //! compute the alignment score
    score_t align();
    
    //! do the trace back through the alignment matrix
    //! with partial recomputation
    //! pre: call align() to fill the top-level matrix
    void trace();
    
    //! set the restriction on the alignment,
    //! mainly used for the k-best algorithm
    void set_restriction(const AlignerRestriction &r_) {r=r_;}
    
    //! return the current restriction,
    //! mainly used for the k-best algorithm
    AlignerRestriction &get_restriction() {return r;}
};

#endif
